""" Python operators """
from typing import Any as Any_
from functools import partial, reduce
import operator

from broqer import Publisher
from broqer.operator_overloading import MapUnary
from .combine_latest import CombineLatest


class Str(MapUnary):
    """ Implementing the functionality of str() for publishers.

    Usage:
    >>> from broqer import op, Value
    >>> num = Value(0)
    >>> literal = op.Str(num)
    >>> literal.get()
    '0'
    """
    def __init__(self, publisher: Publisher) -> None:
        MapUnary.__init__(self, publisher, str)


class Bool(MapUnary):
    """ Implementing the functionality of bool() for publishers. """
    def __init__(self, publisher: Publisher) -> None:
        MapUnary.__init__(self, publisher, bool)


class Not(MapUnary):
    """ Implementing the functionality of not for publishers. """
    def __init__(self, publisher: Publisher) -> None:
        MapUnary.__init__(self, publisher, operator.not_)


class Int(MapUnary):
    """ Implementing the functionality of int() for publishers. """
    def __init__(self, publisher: Publisher) -> None:
        MapUnary.__init__(self, publisher, int)


class Float(MapUnary):
    """ Implementing the functionality of float() for publishers. """
    def __init__(self, publisher: Publisher) -> None:
        MapUnary.__init__(self, publisher, float)


class Repr(MapUnary):
    """ Implementing the functionality of repr() for publishers. """
    def __init__(self, publisher: Publisher) -> None:
        MapUnary.__init__(self, publisher, repr)


class Len(MapUnary):
    """ Implementing the functionality of len() for publishers. """
    def __init__(self, publisher: Publisher) -> None:
        MapUnary.__init__(self, publisher, len)


def _in(item, container):
    return item in container


class In(CombineLatest):
    """ Implementing the functionality of ``in`` operator for publishers.
    :param item: publisher or constant to check for availability in container.
    :param container: container (publisher or constant)
    """
    def __init__(self, item: Any_, container: Any_) -> None:
        if isinstance(item, Publisher) and isinstance(container, Publisher):
            CombineLatest.__init__(self, item, container, map_=_in)
        elif isinstance(item, Publisher):
            function = partial(_in, container=container)
            CombineLatest.__init__(self, item, map_=function)
        else:
            if not isinstance(container, Publisher):
                raise TypeError('Item or container has to be a publisher')
            function = partial(_in, item)
            CombineLatest.__init__(self, container, map_=function)


def _all(*items):
    return all(items)


class All(CombineLatest):
    """ Implement the functionality of ``all`` operator for publishers.
    One big difference is that ``all`` takes an iterator and ``All`` take a
    variable amount of publishers as arguments.
    :param publishers: Publishers evaluated for all to be True
    """
    def __init__(self, *publishers: Any_) -> None:
        CombineLatest.__init__(self, *publishers, map_=_all)


def _any(*items):
    return any(items)


class Any(CombineLatest):
    """ Implement the functionality of ``any`` operator for publishers.
    One big difference is that ``any`` takes an iterator and ``Any`` take a
    variable amount of publishers as arguments.
    :param publishers: Publishers evaluated for one to be True
    """
    def __init__(self, *publishers: Any_) -> None:
        CombineLatest.__init__(self, *publishers, map_=_any)


def _bitwise_or(*items):
    return reduce(operator.or_, items)


class BitwiseOr(CombineLatest):
    """ Implement the functionality of bitwise or (``|``) operator for
    publishers.
    :param publishers: Publishers evaluated for bitwise or
    """
    def __init__(self, *publishers: Any_) -> None:
        CombineLatest.__init__(self, *publishers, map_=_bitwise_or)


def _bitwise_and(*items):
    return reduce(operator.and_, items)


class BitwiseAnd(CombineLatest):
    """ Implement the functionality of bitwise and (``&``) operator for
    publishers.
    :param publishers: Publishers evaluated for bitwise and
    """
    def __init__(self, *publishers: Any_) -> None:
        CombineLatest.__init__(self, *publishers, map_=_bitwise_and)



""" Implementing the Subscriber class """
from abc import ABCMeta, abstractmethod
from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from broqer import Publisher


class Subscriber(metaclass=ABCMeta):  # pylint: disable=too-few-public-methods
    """ A Subscriber is listening to changes of a publisher. As soon as the
    publisher is emitting a value .emit(value) will be called.
    """

    @abstractmethod
    def emit(self, value: Any, who: 'Publisher') -> None:
        """ Send new value to the subscriber
        :param value: value to be send
        :param who: reference to which publisher is emitting
        """



"""
>>> from broqer import Value, Sink, op
>>> s1 = Value()
>>> s2 = Value()

>>> combination = op.CombineLatest(s1, s2)
>>> disposable = combination.subscribe(Sink(print))

CombineLatest is only emitting, when all values are collected:

>>> s1.emit(1)
>>> s2.emit(2)
(1, 2)
>>> s2.emit(3)
(1, 3)

Subscribing to a CombineLatest with all values available is emitting the values
immediate on subscription:

>>> combination.subscribe(Sink(print, 'Second sink:'))
Second sink: (1, 3)
<...>

"""
from functools import wraps
from typing import Any, Dict, MutableSequence, Callable  # noqa: F401

from broqer import Publisher, Subscriber, NONE

from broqer.operator import MultiOperator


class CombineLatest(MultiOperator):
    """ Combine the latest emit of multiple publishers and emit the combination

    :param publishers: source publishers
    :param map_: optional function to be called for evaluation of current state
    :param emit_on: publisher or list of publishers - only emitting result when
        emit comes from one of this list. If None, emit on any source
        publisher.
    """
    def __init__(self, *publishers: Publisher, map_: Callable[..., Any] = None,
                 emit_on=None) -> None:
        MultiOperator.__init__(self, *publishers)

        # ._partial_state is a list keeping the latest emitted values from
        # each publisher. Stateless publishers will always keep the NONE entry
        # in this list.
        self._partial_state = [
            NONE for _ in publishers]  # type: MutableSequence[Any]

        # ._missing is keeping a set of source publishers which are required to
        # emit a value. This set starts with all source publishers. Stateful
        # publishers are required, stateless publishers not (will be removed
        # in .subscribe).
        self._missing = set(publishers)

        # ._index is a lookup table to get the list index based on publisher
        self._index = \
            {p: i for i, p in enumerate(publishers)
             }  # type: Dict[Publisher, int]

        # .emit_on is a set of publishers. When a source publisher is emitting
        # and is not in this set the CombineLatest will not emit a value.
        # If emit_on is None all the publishers will be in the set.
        if isinstance(emit_on, Publisher):
            self._emit_on = (emit_on,)
        else:
            self._emit_on = emit_on

        self._map = map_

    def unsubscribe(self, subscriber: Subscriber) -> None:
        MultiOperator.unsubscribe(self, subscriber)
        if not self._subscriptions:
            self._missing = set(self._orginators)
            self._partial_state[:] = [NONE for _ in self._partial_state]

    def get(self):
        if self._subscriptions:
            return self._state

        values = tuple(p.get() for p in self._orginators)

        if NONE in values:
            return NONE

        if not self._map:
            return tuple(values)

        return self._map(*values)

    def emit(self, value: Any, who: Publisher) -> None:
        if all(who is not p for p in self._orginators):
            raise ValueError('Emit from non assigned publisher')

        # remove source publisher from ._missing
        self._missing.discard(who)

        index = self._index[who]

        # remember state of this source
        self._partial_state[index] = value

        # if emits from publishers are missing or source of this emit
        # is not one of emit_on -> don't evaluate and notify subscribers

        if self._missing or (self._emit_on is not None and all(
                who is not p for p in self._emit_on)):
            return None

        # evaluate
        if self._map:
            state = self._map(*self._partial_state)
        else:
            state = tuple(self._partial_state)

        # if result of _map() was NONE don't emit
        if state is NONE:
            return None

        self._state = state

        return Publisher.notify(self, state)


def build_combine_latest(map_: Callable[..., Any] = None, *, emit_on=None):
    """ Decorator to wrap a function to return a CombineLatest operator.

    :param emit_on: publisher or list of publishers - only emitting result when
        emit comes from one of this list. If None, emit on any source
        publisher.
    """
    def _build_combine_latest(map_: Callable[..., Any]):
        @wraps(map_)
        def _wrapper(*publishers) -> CombineLatest:
            return CombineLatest(*publishers, map_=map_, emit_on=emit_on)
        return _wrapper

    if map_:
        return _build_combine_latest(map_)

    return _build_combine_latest



import asyncio

from broqer import op
from broqer.subject import Subject

value = Subject()


async def main():
    print('Value: ', await (value | op.OnEmitFuture()))
    print('Value: ', await value)
    print('Number of subscribers: %d' % len(value.subscriptions))

loop = asyncio.get_event_loop()

loop.call_later(0.2, value.emit, 1)
loop.call_later(0.4, value.emit, 2)

loop.run_until_complete(main())



""" Module implementing Operator, MultiOperator.
"""
import typing
from abc import abstractmethod, ABCMeta

# pylint: disable=cyclic-import
from broqer import Publisher, SubscriptionDisposable, Subscriber
from broqer.publisher import TValue


class OperatorMeta(ABCMeta):
    """ Metaclass for operators. Defining __ror__ for operator classes.
    This is e.g. used for EvalTrue: p | EvalTrue is equal to EvalTrue(p)
    """
    def __ror__(cls, publisher: Publisher) -> Publisher:
        return cls(publisher)  # pylint: disable=no-value-for-parameter


class Operator(Publisher, Subscriber, metaclass=OperatorMeta):
    """ Base class for operators depending on a single publisher. This
    publisher will be subscribed as soon as this operator is subscribed the
    first time.

    On unsubscription of the last subscriber the dependent publisher will also
    be unsubscripted.
    """
    def __init__(self, publisher: Publisher) -> None:
        Publisher.__init__(self)
        Subscriber.__init__(self)
        self._orginator = publisher
        self.add_dependencies(publisher)

    def subscribe(self, subscriber: 'Subscriber',
                  prepend: bool = False) -> SubscriptionDisposable:
        disposable = Publisher.subscribe(self, subscriber, prepend)

        if len(self._subscriptions) == 1:  # if this was the first subscription
            self._orginator.subscribe(self)

        return disposable

    def unsubscribe(self, subscriber: Subscriber) -> None:
        Publisher.unsubscribe(self, subscriber)

        if not self._subscriptions:
            self._orginator.unsubscribe(self)
            Publisher.reset_state(self)

    def notify(self, value: TValue) -> None:
        raise ValueError('Operator doesn\'t support .notify()')

    def reset_state(self, value: TValue = None) -> None:
        raise ValueError('Operator doesn\'t support .reset_state()')

    @abstractmethod
    def emit(self, value: typing.Any, who: Publisher) -> None:
        """ Send new value to the operator
        :param value: value to be send
        :param who: reference to which publisher is emitting
        """


class OperatorFactory:
    """ OperatorFactory is returning operator objects applied to a publisher
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def apply(self, publisher: Publisher) -> Publisher:
        """ Build a operator applied to the given publisher """

    def __ror__(self, publisher: Publisher) -> Publisher:
        return self.apply(publisher)


class MultiOperator(Publisher, Subscriber):
    """ Base class for operators depending on multiple publishers. Like
    Operator all publishers will be subscribed on first subscription to this
    operator. Accordingly all publishers get unsubscribed on unsubscription
    of the last subscriber.
    """
    def __init__(self, *publishers: Publisher) -> None:
        Publisher.__init__(self)
        Subscriber.__init__(self)
        self._orginators = publishers
        self.add_dependencies(*publishers)

    def subscribe(self, subscriber: 'Subscriber',
                  prepend: bool = False) -> SubscriptionDisposable:
        disposable = Publisher.subscribe(self, subscriber, prepend)

        if len(self._subscriptions) == 1:  # if this was the first subscription
            for publisher in self._orginators:
                # subscribe to all dependent publishers
                publisher.subscribe(self)

        return disposable

    def unsubscribe(self, subscriber: Subscriber) -> None:
        Publisher.unsubscribe(self, subscriber)
        if not self._subscriptions:
            for publisher in self._orginators:
                publisher.unsubscribe(self)
            Publisher.reset_state(self)

    def notify(self, value: TValue) -> None:
        raise ValueError('Operator doesn\'t support .notify()')

    def reset_state(self, value: TValue = None) -> None:
        raise ValueError('Operator doesn\'t support .reset_state()')

    @abstractmethod
    def emit(self, value: typing.Any, who: Publisher) -> None:
        """ Send new value to the operator
        :param value: value to be send
        :param who: reference to which publisher is emitting
        """



from unittest import mock
import pytest

from broqer import Value, Publisher, Subscriber, Sink, op
import operator

def test_operator_with_publishers():
    v1 = Value(0)
    v2 = Value(0)

    o = v1 + v2

    assert isinstance(o, Publisher)
    assert isinstance(o, Subscriber)
    assert o.get() == 0

    v1.emit(1)
    assert o.get() == 1

    assert len(o.subscriptions) == 0

    mock_sink = mock.Mock()

    o.subscribe(Sink(mock_sink))
    assert len(o.subscriptions) == 1
    mock_sink.assert_called_once_with(1)
    mock_sink.reset_mock()

    v2.emit(3)
    mock_sink.assert_called_once_with(4)

    with pytest.raises(ValueError):
        o.emit(0, who=Publisher())

    with pytest.raises(ValueError):
        Value(1).subscribe(o)

def test_operator_with_constant():
    v1 = Value(0)
    v2 = 1

    o = v1 + v2

    assert isinstance(o, Publisher)
    assert o.get() == 1

    v1.emit(1)
    assert o.get() == 2

    assert len(o.subscriptions) == 0

    mock_sink = mock.Mock()

    o.subscribe(Sink(mock_sink))
    assert len(o.subscriptions) == 1
    mock_sink.assert_called_once_with(2)
    mock_sink.reset_mock()

    v1.emit(3)
    mock_sink.assert_called_once_with(4)

    with pytest.raises(TypeError):
        Value(1) | o

    with pytest.raises(ValueError):
        o.emit(0, who=Publisher())

def test_operator_with_constant_r():
    v1 = 1
    v2 = Value(0)

    o = v1 - v2

    assert isinstance(o, Publisher)
    assert o.get() == 1

    v2.emit(1)
    assert o.get() == 0

    assert len(o.subscriptions) == 0

    mock_sink = mock.Mock()

    o.subscribe(Sink(mock_sink))
    assert len(o.subscriptions) == 1
    mock_sink.assert_called_once_with(0)
    mock_sink.reset_mock()

    v2.emit(3)
    mock_sink.assert_called_once_with(-2)

    with pytest.raises(TypeError):
        Value(1) | o

    with pytest.raises(ValueError):
        o.emit(0, who=Publisher())

@pytest.mark.parametrize('operator, l_value, r_value, result', [
    (operator.lt, 0, 1, True), (operator.lt, 1, 0, False), (operator.lt, 1, 1, False), (operator.lt, 1, 'foo', TypeError),
    (operator.le, 0, 1, True), (operator.le, 1, 0, False), (operator.le, 1, 1, True), (operator.le, 1, 'foo', TypeError),
    (operator.eq, 1, 0, False), (operator.eq, 1, 1, True), (operator.eq, 1, 'foo', False), (operator.eq, 'foo', 'foo', True),
    (operator.ne, 1, 0, True), (operator.ne, 1, 1, False), (operator.ne, 1, 'foo', True), (operator.ne, 'foo', 'foo', False),
    (operator.ge, 0, 1, False), (operator.ge, 1, 0, True), (operator.ge, 1, 1, True), (operator.ge, 1, 'foo', TypeError),
    (operator.gt, 0, 1, False), (operator.gt, 1, 0, True), (operator.gt, 1, 1, False), (operator.gt, 1, 'foo', TypeError),
    (operator.add, 0, 1, 1), (operator.add, 0, 1.1, 1.1), (operator.add, 'ab', 'cd', 'abcd'), (operator.add, 'ab', 1, TypeError),
    (operator.and_, 0, 0, 0), (operator.and_, 5, 1, 1), (operator.and_, 5, 4, 4), (operator.and_, 'ab', 1, TypeError),
    (operator.lshift, 0, 0, 0), (operator.lshift, 5, 1, 10), (operator.lshift, 5, 4, 80), (operator.lshift, 'ab', 1, TypeError),
    (operator.mod, 0, 0, ZeroDivisionError), (operator.mod, 5, 1, 0), (operator.mod, 5, 4, 1), (operator.mod, 1, 'ab', TypeError),
    # str%tuple is tested seperatly
    (operator.mul, 0, 0, 0), (operator.mul, 5, 1, 5), (operator.mul, 'ab', 4, 'abababab'), (operator.mul, 1, 'ab', 'ab'), (operator.mul, (1,2), 'ab', TypeError),
    (operator.pow, 0, 0, 1), (operator.pow, 5, 2, 25), (operator.pow, 'ab', 4, TypeError),
    (operator.rshift, 0, 0, 0), (operator.rshift, 5, 1, 2), (operator.rshift, 5, 4, 0), (operator.rshift, 'ab', 1, TypeError),
    (operator.sub, 0, 0, 0), (operator.sub, 5, 1, 4), (operator.sub, 1, 4, -3), (operator.sub, 'ab', 1, TypeError),
    (operator.xor, 0, 0, 0), (operator.xor, 5, 1, 4), (operator.xor, 5, 3, 6), (operator.xor, 'ab', 1, TypeError),
    # concat is tested seperatly
    # getitem is tested seperatly
    (operator.floordiv, 0, 0, ZeroDivisionError), (operator.floordiv, 5, 4, 1), (operator.floordiv, 5, 'ab', TypeError),
    (operator.truediv, 0, 0, ZeroDivisionError), (operator.truediv, 5, 4, 1.25), (operator.truediv, 5, 'ab', TypeError),
])
def test_with_publisher(operator, l_value, r_value, result):
    vl = Value(l_value)
    vr = Value(r_value)
    cl = l_value
    cr = r_value

    o1 = operator(vl, vr)
    o2 = operator(vl, cr)
    o3 = operator(cl, vr)
    try:
        o4 = operator(cl, cr)
    except Exception as e:
        assert isinstance(e, result)
        o4 = result  # to pass the following test

    mock_sink_o3 = mock.Mock()

    try:
        o3.subscribe(Sink(mock_sink_o3))
    except Exception as e:
        assert isinstance(e, result)

    for output in (o1, o2, o3):
        assert isinstance(output, Publisher)

    assert o4 == result

    for output in (o1, o2, o3):
        try:
            assert output.get() == result
        except Exception as e:
            assert isinstance(e, result)

def test_wrong_comparision():
    p1 = Publisher()
    p2 = Publisher()

    with pytest.raises(ValueError):
        assert p1 == p2

    with pytest.raises(ValueError):
        if p1 == p2: pass

    with pytest.raises(ValueError):
        if p1 != p2: pass

    with pytest.raises(ValueError):
        assert p2 in (p1, p2)

    with pytest.raises(ValueError):
        p1 in (p2, p2)

    with pytest.raises(ValueError):
        assert p1 not in (p2, p2)

    l = [p1, p2]
    with pytest.raises(ValueError):
        l.remove(p2)

def test_mod_str():
    v1 = Publisher('%.2f %d')
    v2 = Value((0,0))

    o = v1%v2

    assert isinstance(o, Publisher)
    assert o.get() == '0.00 0'

    v2.emit((1,3))
    assert o.get() == '1.00 3'

def test_concat():
    v1 = Publisher((1,2))
    v2 = Value((0,0))

    o = v1+v2

    assert isinstance(o, Publisher)
    assert o.get() == (1,2,0,0)

    v2.emit((1,3))
    assert o.get() == (1,2,1,3)

def test_getitem():
    v1 = Value(('a','b','c'))
    v2 = Value(2)

    o = (v1[v2])

    assert isinstance(o, Publisher)
    assert o.get() == 'c'

    v2.emit(1)
    assert o.get() == 'b'

from collections import Counter
import math

@pytest.mark.parametrize('operator, value, result', [
    (operator.neg, -5, 5), (operator.neg, 'ab', TypeError), (operator.pos, -5, -5),
    (operator.pos, Counter({'a':0, 'b':1}), Counter({'b':1})), (operator.abs, -5, 5),
    (operator.invert, 5, -6), (round, 5.2, 5), (round, 5.8, 6), (math.trunc, -5.2, -5),
    (math.floor, -5.2, -6), (math.ceil, -5.2, -5),
    (op.Not, 5, False), (op.Not, 0, True), (op.Not, True, False),
    (op.Not, False, True),
    (op.Str, 123, '123'), (op.Str, (1, 2), '(1, 2)'), (op.Str, 1.23, '1.23'),
    (op.Bool, 0, False), (op.Bool, 1, True), (op.Bool, 'False', True), (op.Bool, (1,2,3), True),
    (op.Bool, {}, False), (op.Bool, None, False), (op.Bool, 513.17, True), (op.Bool, 0.0, False),
    (op.Int, 1.99, 1), (op.Int, '123', 123), (op.Int, (1, 2, 3), TypeError), (op.Int, None, TypeError),
    (op.Float, 1, 1.0), (op.Float, '12.34', 12.34), (op.Float, 'abc', ValueError),
    (op.Repr, 123, '123'), (op.Repr, 'abc', '\'abc\''), (op.Repr, int, '<class \'int\'>'),
    (op.Len, (), 0), (op.Len, [1,2,3], 3), (op.Len, 'abcde', 5), (op.Len, None, TypeError),
])
def test_unary_operators(operator, value, result):
    v = Publisher(value)

    try:
        value_applied = operator(v).get()
    except Exception as e:
        assert isinstance(e, result)
    else:
        assert value_applied == result

    cb = mock.Mock()

    try:
        operator(v).subscribe(Sink(cb))
    except Exception as e:
        assert isinstance(e, result)
    else:
        assert cb.mock_called_once_with(result)

    with pytest.raises(TypeError):
        Value(1) | operator(v)

    with pytest.raises(ValueError):
        operator(v).emit(0, who=Publisher())

def test_in_operator():
    pi = Value(1)
    ci = 1

    pc = Value((1,2,3))
    cc = (1, 2, 3)

    dut1 = op.In(pi, pc)
    dut2 = op.In(ci, pc)
    dut3 = op.In(pi, cc)
    with pytest.raises(TypeError):
        op.In(ci, cc)

    assert dut1.get() == True
    assert dut2.get() == True
    assert dut3.get() == True

    pi.emit('a')
    pc.emit((2.3, 'b'))

    assert dut1.get() == False
    assert dut2.get() == False
    assert dut3.get() == False

def test_getattr_method():
    p = Publisher('')
    p.inherit_type(str)

    dut1 = p.split()
    dut2 = p.split(',')
    dut3 = p.split(sep = '!')

    mock1 = mock.Mock()
    mock2 = mock.Mock()
    mock3 = mock.Mock()

    dut1.subscribe(Sink(mock1))
    dut2.subscribe(Sink(mock2))
    dut3.subscribe(Sink(mock3))

    assert dut1.get() == []
    assert dut2.get() == ['']
    assert dut3.get() == ['']

    mock1.assert_called_once_with([])
    mock2.assert_called_once_with([''])
    mock3.assert_called_once_with([''])

    mock1.reset_mock()
    mock2.reset_mock()
    mock3.reset_mock()

    p.notify('This is just a test, honestly!')

    assert dut1.get() == ['This', 'is', 'just', 'a', 'test,', 'honestly!']
    assert dut2.get() == ['This is just a test', ' honestly!']
    assert dut3.get() == ['This is just a test, honestly', '']

    mock1.assert_called_once_with(['This', 'is', 'just', 'a', 'test,', 'honestly!'])
    mock2.assert_called_once_with(['This is just a test', ' honestly!'])
    mock3.assert_called_once_with(['This is just a test, honestly', ''])

def test_inherit_getattr():
    p = Publisher('')
    p.inherit_type(str)

    dut = p.lower().split(' ')
    m = mock.Mock()
    dut.subscribe(Sink(m))
    m.assert_called_once_with([''])
    m.reset_mock()

    p.notify('This is a TEST')
    m.assert_called_once_with(['this', 'is', 'a', 'test'])

def test_inherit_with_operators():
    p = Publisher('')
    p.inherit_type(str)

    dut = op.Len(('abc' + p + 'ghi').upper())
    m = mock.Mock()
    dut.subscribe(Sink(m))
    m.assert_called_once_with(6)
    m.reset_mock()

    p.notify('def')
    m.assert_called_once_with(9)

def test_getattr_attribute():
    class Foo:
        a = None

        def __init__(self, a=5):
            self.a = a

    p = Publisher(Foo(3))
    p.inherit_type(Foo)

    dut = p.a
    m = mock.Mock()
    dut.subscribe(Sink(m))

    m.assert_called_once_with(3)
    assert dut.get() == 3
    m.reset_mock()

    p.notify(Foo(4))

    assert dut.get() == 4

    m.assert_called_once_with(4)

    with pytest.raises(ValueError):
        dut.emit(0, who=Publisher())

    with pytest.raises(AttributeError):
        dut.assnign(5)


def test_getattr_without_inherit():
    p = Publisher()

    class Foo:
        a = None

        def __init__(self, a=5):
            self.a = a

    with pytest.raises(AttributeError):
        dut = p.a

    with pytest.raises(AttributeError):
        p.assnign(5)

@pytest.mark.parametrize('operator, values, result', [
    (op.All, (False, False, False), False),
    (op.All, (False, True, False), False),
    (op.All, (True, True, True), True),
    (op.Any, (False, False, False), False),
    (op.Any, (False, True, False), True),
    (op.Any, (True, True, True), True),
    (op.BitwiseAnd, (0, 5, 15), 0),
    (op.BitwiseAnd, (7, 14, 255), 6),
    (op.BitwiseAnd, (3,), 3),
    (op.BitwiseOr, (0, 5, 8), 13),
    (op.BitwiseOr, (7, 14, 255), 255),
    (op.BitwiseOr, (3,), 3),
])
def test_multi_operators(operator, values, result):
    sources = [Publisher(v) for v in values]
    dut = operator(*sources)
    assert dut.get() == result



"""
Filters values based on a ``predicate`` function

Usage:

>>> from broqer import Value, op, Sink
>>> s = Value()

>>> filtered_publisher = s | op.Filter(lambda v:v>0)
>>> _disposable = filtered_publisher.subscribe(Sink(print))

>>> s.emit(1)
1
>>> s.emit(-1)
>>> s.emit(0)
>>> _disposable.dispose()

Also possible with additional args and kwargs:

>>> import operator
>>> filtered_publisher = s | op.Filter(operator.and_, 0x01)
>>> _disposable = filtered_publisher.subscribe(Sink(print))
>>> s.emit(100)
>>> s.emit(101)
101

"""
from functools import partial, wraps
from typing import Any, Callable

from broqer import NONE, Publisher
from broqer.operator import Operator, OperatorFactory, OperatorMeta


class AppliedFilter(Operator):
    """ Filter object applied to publisher (see Filter) """
    def __init__(self, publisher: Publisher, predicate: Callable[[Any], bool],
                 unpack: bool = False) -> None:
        Operator.__init__(self, publisher)
        self._predicate = predicate
        self._unpack = unpack

    def get(self) -> Any:
        if self._subscriptions:
            return self._state

        value = self._orginator.get()  # type: Any

        if self._unpack:
            # assert isinstance(value, (list, tuple))
            if self._predicate(*value):
                return value

        elif self._predicate(value):
            return value

        return NONE

    def emit(self, value: Any, who: Publisher) -> None:
        if who is not self._orginator:
            raise ValueError('Emit from non assigned publisher')

        if self._unpack:
            if self._predicate(*value):
                return Publisher.notify(self, value)
        elif self._predicate(value):
            return Publisher.notify(self, value)
        return None


class Filter(OperatorFactory):  # pylint: disable=too-few-public-methods
    """ Filters values based on a ``predicate`` function
    :param predicate: function to evaluate the filtering
    :param \\*args: variable arguments to be used for evaluating predicate
    :param unpack: value from emits will be unpacked (\\*value)
    :param \\*\\*kwargs: keyword arguments to be used for evaluating predicate
    """
    def __init__(self, predicate: Callable[[Any], bool],
                 *args, unpack: bool = False, **kwargs) -> None:
        self._predicate = partial(predicate, *args, **kwargs)  # type: Callable
        self._unpack = unpack

    def apply(self, publisher: Publisher):
        return AppliedFilter(publisher, self._predicate, self._unpack)


class EvalTrue(Operator, metaclass=OperatorMeta):
    """ Emits all values which evaluates for True.

    This operator can be used in the pipline style (v | EvalTrue) or as
    standalone operation (EvalTrue(v)).
    """
    def __init__(self, publisher: Publisher) -> None:
        Operator.__init__(self, publisher)

    def get(self) -> Any:
        if self._subscriptions:
            return self._state

        assert isinstance(self._orginator, Publisher)

        value = self._orginator.get()  # type: Any

        if bool(value):
            return value

        return NONE

    def emit(self, value: Any, who: Publisher) -> None:
        if who is not self._orginator:
            raise ValueError('Emit from non assigned publisher')

        if bool(value):
            return Publisher.notify(self, value)
        return None


class EvalFalse(Operator, metaclass=OperatorMeta):
    """ Filters all emits which evaluates for False.

    This operator can be used in the pipline style (v | EvalFalse or as
    standalone operation (EvalFalse(v))."""
    def __init__(self, publisher: Publisher) -> None:
        Operator.__init__(self, publisher)

    def get(self) -> Any:
        if self._subscriptions:
            return self._state

        assert isinstance(self._orginator, Publisher)

        value = self._orginator.get()  # type: Any

        if not bool(value):
            return value

        return NONE

    def emit(self, value: Any, who: Publisher) -> None:
        if who is not self._orginator:
            raise ValueError('Emit from non assigned publisher')

        if not bool(value):
            return Publisher.notify(self, value)
        return None


def build_filter(predicate: Callable[[Any], bool] = None, *,
                 unpack: bool = False):
    """ Decorator to wrap a function to return a Filter operator.

    :param function: function to be wrapped
    :param unpack: value from emits will be unpacked (*value)
    """
    def _build_filter(predicate):
        return Filter(predicate, unpack=unpack)

    if predicate:
        return _build_filter(predicate)

    return _build_filter


def build_filter_factory(predicate: Callable[[Any], bool] = None, *,
                         unpack: bool = False):
    """ Decorator to wrap a function to return a factory for Filter operators.

    :param predicate: function to be wrapped
    :param unpack: value from emits will be unpacked (*value)
    """
    def _build_filter(predicate: Callable[[Any], bool]):
        @wraps(predicate)
        def _wrapper(*args, **kwargs) -> Filter:
            if 'unpack' in kwargs:
                raise TypeError('"unpack" has to be defined by decorator')
            return Filter(predicate, *args, unpack=unpack, **kwargs)
        return _wrapper

    if predicate:
        return _build_filter(predicate)

    return _build_filter



from unittest import mock

from broqer import Disposable, SubscriptionDisposable, Publisher, Value


def test_disposable():
    """ Testing base class Disposable
    """
    m = mock.Mock()

    class MyDisposable(Disposable):
        def dispose(self):
            m('disposed')

    d = MyDisposable()

    # test simple .dispose() call
    m.assert_not_called()
    d.dispose()
    m.assert_called_once_with('disposed')

    m.reset_mock()

    # test context manager
    with d:
        m.assert_not_called()

    m.assert_called_once_with('disposed')

    m.reset_mock()

    # test context manager creating the disposable
    with MyDisposable() as d2:
        m.assert_not_called()

    m.assert_called_once_with('disposed')


def test_subscription_disposable():
    p = Publisher()
    v = Value(0)

    assert len(p.subscriptions) == 0

    disposable = p.subscribe(v)

    assert len(p.subscriptions) == 1

    assert disposable.publisher is p
    assert disposable.subscriber is v

    disposable.dispose()

    assert len(p.subscriptions) == 0

    with p.subscribe(v):
        assert len(p.subscriptions) == 1

    assert len(p.subscriptions) == 0






import inspect
import traceback
import logging
import threading
import collections
import heapq
import asyncio
from asyncio import TimerHandle

log = logging.getLogger(__name__)


def _format_handle(handle):
    cb = handle._callback
    if inspect.ismethod(cb) and isinstance(cb.__self__, asyncio.Task):
        # format the task
        return repr(cb.__self__)
    else:
        return str(handle)


def _run_until_complete_cb(fut):
    exc = fut._exception
    if (isinstance(exc, BaseException) and not isinstance(exc, Exception)):
        # Issue #22429: run_forever() already finished, no need to
        # stop it.
        return
    fut._loop.stop()


def isfuture(obj):
    """ Check for a Future.
    This returns True when obj is a Future instance or is advertising
    itself as duck-type compatible by setting _asyncio_future_blocking.
    See comment in Future for more details.
    """
    return getattr(obj, '_asyncio_future_blocking', None) is not None


class VirtualTimeEventLoop(asyncio.AbstractEventLoop):

    def __init__(self):
        self._timer_cancelled_count = 0
        self._closed = False
        self._stopping = False
        self._ready = collections.deque()
        self._scheduled = []
        self._thread_id = None
        self._exception_handler = None
        self._current_handle = None
        self._debug = False
        self._time = 0

    def run_forever(self):
        """ Run until stop() is called."""
        self._check_closed()

        if self.is_running():
            raise RuntimeError('Event loop is running.')

        self._thread_id = threading.get_ident()
        try:
            while True:
                self._run_once()
                if self._stopping:
                    break
        finally:
            self._stopping = False
            self._thread_id = None

    def run_until_complete(self, future):
        """ Run the event loop until a Future is done.
        Return the Future's result, or raise its exception.
        """
        self._check_closed()

        new_task = not isfuture(future)
        future = asyncio.ensure_future(future, loop=self)
        if new_task:
            # An exception is raised if the future didn't complete, so there
            # is no need to log the 'destroy pending task' message
            future._log_destroy_pending = False

        future.add_done_callback(_run_until_complete_cb)
        try:
            self.run_forever()
        except:
            if new_task and future.done() and not future.cancelled():
                # The coroutine raised a BaseException. Consume the exception
                # to not log a warning, the caller doesn't have access to the
                # local task.
                future.exception()
            raise
        future.remove_done_callback(_run_until_complete_cb)
        if not future.done():
            raise RuntimeError('Event loop stopped before Future completed.')

        return future.result()

    def stop(self):
        """ Stop running the event loop.
        Every callback already scheduled will still run.  This simply informs
        run_forever to stop looping after a complete iteration.
        """
        self._stopping = True

    def close(self):
        """ Close the event loop.
        This clears the queues and shuts down the executor,
        but does not wait for the executor to finish.
        The event loop must not be running.
        """
        if self.is_running():
            raise RuntimeError('Cannot close a running event loop')
        if self._closed:
            return

        self._closed = True
        self._ready.clear()
        self._scheduled.clear()

    def is_closed(self):
        """ Returns True if the event loop was closed."""
        return self._closed

    def is_running(self):
        """ Returns True if the event loop is running."""
        return (self._thread_id is not None)

    def call_later(self, delay, callback, *args):
        timer = self.call_at(self.time() + delay, callback, *args)
        if timer._source_traceback:
            del timer._source_traceback[-1]
        return timer

    def call_at(self, when, callback, *args):
        timer = TimerHandle(when, callback, args, self)
        if timer._source_traceback:
            del timer._source_traceback[-1]
        heapq.heappush(self._scheduled, timer)
        timer._scheduled = True
        return timer

    def call_soon(self, callback, *args, context=None):
        handle = self._call_soon(callback, args)
        if handle._source_traceback:
            del handle._source_traceback[-1]
        return handle

    def _call_soon(self, callback, args):
        if (asyncio.iscoroutine(callback) or asyncio.iscoroutinefunction(callback)):
            raise TypeError('coroutines cannot be used with call_soon()')
        self._check_closed()
        handle = asyncio.Handle(callback, args, self)
        if handle._source_traceback:
            del handle._source_traceback[-1]
        self._ready.append(handle)
        return handle

    def call_soon_threadsafe(self, callback, *args):
        raise NotImplementedError

    def time(self):
        return self._time

    def create_future(self):
        """ Create a Future object attached to the loop."""
        return asyncio.Future(loop=self)

    def create_task(self, coro):
        """ Schedule a coroutine object.
        Return a task object."""

        self._check_closed()
        task = asyncio.Task(coro, loop=self)
        if task._source_traceback:
            del task._source_traceback[-1]
        return task

    def default_exception_handler(self, context):
        """ Default exception handler.
        This is called when an exception occurs and no exception
        handler is set, and can be called by a custom exception
        handler that wants to defer to the default behavior.
        The context parameter has the same meaning as in
        `call_exception_handler()`."""

        message = context.get('message')
        if not message:
            message = 'Unhandled exception in event loop'

        exception = context.get('exception')
        if exception is not None:
            exc_info = (type(exception), exception, exception.__traceback__)
        else:
            exc_info = False

        if ('source_traceback' not in context
        and self._current_handle is not None
        and self._current_handle._source_traceback):
            context['handle_traceback'] = self._current_handle._source_traceback

        log_lines = [message]
        for key in sorted(context):
            if key in {'message', 'exception'}:
                continue
            value = context[key]
            if key == 'source_traceback':
                tb = ''.join(traceback.format_list(value))
                value = 'Object created at (most recent call last):\n'
                value += tb.rstrip()
            elif key == 'handle_traceback':
                tb = ''.join(traceback.format_list(value))
                value = 'Handle created at (most recent call last):\n'
                value += tb.rstrip()
            else:
                value = repr(value)
            log_lines.append('{}: {}'.format(key, value))

        log.error('\n'.join(log_lines), exc_info=exc_info)

    def call_exception_handler(self, context):
        """ Call the current event loop's exception handler.
        The context argument is a dict containing the following keys:
        - 'message': Error message;
        - 'exception' (optional): Exception object;
        - 'future' (optional): Future instance;
        - 'handle' (optional): Handle instance;
        - 'protocol' (optional): Protocol instance;
        - 'transport' (optional): Transport instance;
        - 'socket' (optional): Socket instance;
        - 'asyncgen' (optional): Asynchronous generator that caused
                                 the exception.
        New keys maybe introduced in the future.
        Note: do not overload this method in an event loop subclass.
        For custom exception handling, use the
        `set_exception_handler()` method.
        """
        if self._exception_handler is None:
            try:
                self.default_exception_handler(context)
            except Exception:
                # Second protection layer for unexpected errors
                # in the default implementation, as well as for subclassed
                # event loops with overloaded 'default_exception_handler'.
                log.error('Exception in default exception handler', exc_info=True)
        else:
            try:
                self._exception_handler(self, context)
            except Exception as exc:
                # Exception in the user set custom exception handler.
                try:
                    # Let's try default handler.
                    self.default_exception_handler({
                        'message': 'Unhandled error in exception handler',
                        'exception': exc,
                        'context': context,
                    })
                except Exception:
                    # Guard 'default_exception_handler' in case it is
                    # overloaded.
                    log.error('Exception in default exception handler '
                              'while handling an unexpected error '
                              'in custom exception handler',
                              exc_info=True)

    def get_debug(self):
        return self._debug

    def _run_once(self):
        # Handle 'later' callbacks that are ready.

        sched_count = len(self._scheduled)
        if (sched_count and self._timer_cancelled_count):
            # Remove delayed calls that were cancelled if their number
            # is too high
            new_scheduled = []
            for handle in self._scheduled:
                if handle._cancelled:
                    handle._scheduled = False
                else:
                    new_scheduled.append(handle)

            heapq.heapify(new_scheduled)
            self._scheduled = new_scheduled
            self._timer_cancelled_count = 0
        else:
            # Remove delayed calls that were cancelled from head of queue.
            while self._scheduled and self._scheduled[0]._cancelled:
                self._timer_cancelled_count -= 1
                handle = heapq.heappop(self._scheduled)
                handle._scheduled = False

        # Process callbacks one at a time and advance time
        if self._scheduled and not self._ready:
            handle = self._scheduled[0]
            handle = heapq.heappop(self._scheduled)
            handle._scheduled = False
            self._time = handle._when
            self._ready.append(handle)

        ntodo = len(self._ready)
        for i in range(ntodo):
            handle = self._ready.popleft()
            if handle._cancelled:
                continue
            if self._debug:
                try:
                    self._current_handle = handle
                    t0 = self.time()
                    handle._run()
                    dt = self.time() - t0
                    if dt >= self.slow_callback_duration:
                        log.warning('Executing %s took %.3f seconds', _format_handle(handle), dt)
                finally:
                    self._current_handle = None
            else:
                handle._run()
        handle = None  # Needed to break cycles when an exception occurs.

    def _check_closed(self):
        if self._closed:
            raise RuntimeError('Event loop is closed')

    def _timer_handle_cancelled(self, handle):
        """ Notification that a TimerHandle has been cancelled."""
        if handle._scheduled:
            self._timer_cancelled_count += 1



""" Implements Trace subscriber """
from time import time
from typing import TYPE_CHECKING, Any, Callable, Optional

from .sink import Sink

if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from broqer import Publisher


class Trace(Sink):
    """ Trace is a subscriber used for debugging purpose. On subscription
    it will use the prepend flag to be the first callback called when the
    publisher of interest is emitting.
    :param callback: optional function to call
    :param \\*args: arguments used additionally when calling callback
    :param unpack: value from emits will be unpacked (\\*value)
    :param label: string to be used on output
    :param \\*\\*kwargs: keyword arguments used when calling callback
    """
    def __init__(self,  # pylint: disable=keyword-arg-before-vararg
                 function: Optional[Callable[..., None]] = None,
                 *args, unpack=False, label=None, **kwargs) -> None:
        Sink.__init__(self, function, *args, unpack=unpack, **kwargs)
        self._label = label

    def emit(self, value: Any, who: 'Publisher'):
        self._trace_handler(who, value, label=self._label)
        Sink.emit(self, value, who=who)

    @classmethod
    def set_handler(cls, handler):
        """ Setting the handler for tracing information """
        cls._trace_handler = handler

    _timestamp_start = time()

    @staticmethod
    def _trace_handler(publisher: 'Publisher', value, label=None):
        """ Default trace handler is printing the timestamp, the publisher name
        and the emitted value
        """
        line = '--- %8.3f: ' % (time() - Trace._timestamp_start)
        line += repr(publisher) if label is None else label
        line += ' %r' % (value,)
        print(line)
